#include "bsdisks.h"
#include "adaptors.h"
#include "objectmanager.h"
#include "block.h"
#include "blockfilesystem.h"
#include "drive.h"
#include "filesystemprober.h"

DBUSManagerStruct ObjectManager::GetManagedObjects()
{
    DBUSManagerStruct r;
    return r;
}


void ObjectManager::initialProbe()
{
    GeomProber prober;

    QObject::connect(&prober, &GeomProber::gotDisk,
        [this](DiskInfo di)
        {
            addDrive(di.devName());

            // assume that any object from `geom disk list` has a block device in /dev
            addBlock(di.devName());
        });
    QObject::connect(&prober, &GeomProber::gotPart,
        [this](PartInfo pi)
        {
            foreach(auto devName, pi.partitions().keys())
                addBlock(devName);
        });

    prober.run();
}

void ObjectManager::geomDevAdded(QString dev)
{
    // don't do anything with stuff like diskid/blabla
    if(dev.contains('/'))
        return;

    auto b = addBlock(dev);

    if(b->bPartition)
        updatePartitions(b->bPartition->partTableBlock);
}


void ObjectManager::filesystemAdded(Block* b, QString fs)
{
    Q_ASSERT(b->bFilesystem == nullptr);

    auto bfs = new BlockFilesystem(b);
    bfs->filesystem = fs;

    b->bFilesystem = bfs;
    b->idUsage = QStringLiteral("filesystem");

    if(fs == QStringLiteral("msdosfs"))
    {
        b->idType = QStringLiteral("vfat");
    }
    else
    {
        b->idType = fs;
    }

    new FilesystemAdaptor(b);
    addInterface(b->dbusPath, "org.freedesktop.UDisks2.Filesystem");
}


void ObjectManager::removeBlock(QString dev)
{
    if(m_blockObjects.contains(dev))
    {
        auto* b = m_blockObjects.take(dev);
        qDebug() << "Unregistering " + b->dbusPath.path();
        if(b->bFilesystem)
            removeInterface(b->dbusPath, "org.freedesktop.UDisks2.Filesystem");
        if(b->bPartition)
            removeInterface(b->dbusPath, "org.freedesktop.UDisks2.Partition");
        if(b->bPartTable)
            removeInterface(b->dbusPath, "org.freedesktop.UDisks2.PartitionTable");
        removeInterface(b->dbusPath, "org.freedesktop.UDisks2.Block");
        QDBusConnection::systemBus().unregisterObject(b->dbusPath.path());
        delete b;
    }
}


Block* ObjectManager::addBlock(QString dev)
{
    // don't do anything with stuff like diskid/blabla
    if(dev.contains('/'))
        return nullptr;


    auto b = m_blockObjects.value(dev);
    Q_ASSERT(!b);

    b = new Block;
    m_blockObjects.insert(dev, b);

    QString devPath = UDisksBlockDevices + dev;
    b->dev = ("/dev/" + dev).toLatin1();
    b->dbusPath = QDBusObjectPath(devPath);

    new BlockAdaptor(b);

    qDebug() << "Registering " + devPath;
    QDBusConnection::systemBus().registerObject(devPath, b);

    addInterface(b->dbusPath, "org.freedesktop.UDisks2.Block");

    startFilesystemProbe(b);
    startGeomProbe(b);

    return b;
}


void ObjectManager::removeDrive(QString dev)
{
    if(m_driveObjects.contains(dev))
    {
        auto* d = m_driveObjects.take(dev);
        qDebug() << "Unregistering " + d->dbusPath.path();
        removeInterface(d->dbusPath, "org.freedesktop.UDisks2.Drive");
        QDBusConnection::systemBus().unregisterObject(d->dbusPath.path());
        delete d;
    }
}


Drive* ObjectManager::addDrive(QString dev)
{
    auto d = m_driveObjects.value(dev);
    Q_ASSERT(!d);

    d = new Drive;
    m_driveObjects.insert(dev, d);

    QString devPath = UDisksDrives + dev;
    d->dbusPath = QDBusObjectPath(devPath);
    d->geomName = dev;
    new DriveAdaptor(d);

    qDebug() << "Registering " + devPath;
    QDBusConnection::systemBus().registerObject(devPath, d);

    addInterface(d->dbusPath, "org.freedesktop.UDisks2.Drive");

    return d;
}

void ObjectManager::startFilesystemProbe(Block* b)
{
    auto prober = new FilesystemProber(b->dev);

    QObject::connect(prober, &FilesystemProber::finished, this,
                     [this, b](QString fs)
    {
        if(!fs.isEmpty())
            filesystemAdded(b, fs);
    });

    QThreadPool::globalInstance()->start(prober);
}


void ObjectManager::startGeomProbe(Block* b)
{
    QFileInfo info(b->dev);
    auto prober = new GeomProber(info.fileName());

    QObject::connect(prober, &GeomProber::gotLabels, this,
                     [this, b](QStringList labels)
    {
        b->labels = labels;
    });
    QObject::connect(prober, &GeomProber::gotDisk, this,
                     [this, b](DiskInfo di)
    {
        b->size = di.mediaSize();
        b->description = di.descr();
        b->identifier = di.ident();
    });
    QObject::connect(prober, &GeomProber::gotPart, this,
                     [this, b](PartInfo pi)
    {
        Q_ASSERT(!b->bPartTable);

        b->bPartTable = new BlockPartTable(b);
        b->bPartTable->setTableType(pi.scheme());
        b->parts = pi.partitions();

        new PartitionTableAdaptor(b);
        addInterface(b->dbusPath, QStringLiteral("org.freedesktop.UDisks2.PartitionTable"));

        updatePartitions(b);
    });

    QThreadPool::globalInstance()->start(prober);
}


void ObjectManager::updatePartitions(Block* table)
{
    for(auto it = m_blockObjects.constBegin(); it != m_blockObjects.cend(); it++)
    {
        QFileInfo info(table->dev);

        if(!it.key().startsWith(info.fileName()))
            continue;
        if(it.key() == info.fileName())
            continue;

        Block* b = it.value();

        auto bp = new BlockPartition(b);
        auto pi = table->parts[it.key()];

        // TODO: find out why there are "da0s1" and "/dev/da0s1" keys
        bp->partTableBlock = table;
        bp->partitionType = pi.type();
        bp->number = pi.index();
        bp->offset = pi.offset();
        bp->size = pi.length();

        b->bPartition = bp;

        new PartitionAdaptor(b);
        addInterface(b->dbusPath, "org.freedesktop.UDisks2.Partition");
    }
    table->parts.clear();
}


void ObjectManager::addInterface(QDBusObjectPath path, QString iface)
{
    QVariantMapMap interfaces;
    interfaces[iface] = QVariantMap();
    emit InterfacesAdded(path, interfaces);
}


void ObjectManager::removeInterface(QDBusObjectPath path, QString iface)
{
    QStringList interfaces;
    interfaces.push_back(iface);
    emit InterfacesRemoved(path, interfaces);
}

